import math
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import interp1d
from scipy.signal import detrend
import scipy.fftpack
import operator

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# BEGIN LOCAL FUNCTIONS
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
def file_len(fname):
    """Return number of lines in a file (1 if file empty)

    Keyword arguments:
    fname -- filename (string)
    """
    i = 0
    with open(fname) as f:
        for i, line in enumerate(f):
            pass
    return i + 1

#%%%%%%%%%%

def ngaussian(x,pos,wid,n):
    """Return shape function for filter.

    Keyword arguments:
    ngaussian(x,pos,wid) = peak centered on x=pos, half-width=wid
    x may be scalar, vector, or matrix, pos and wid both scalar
    Shape is Gaussian when n=1, becomes more rectangular as n increases.
    Example: ngaussian([1 2 3],1,2,1) gives result [1.0000    0.5000    0.0625]
    """
    fn = [0 for j in range(len(x))]
    
    for i in range(len(x)):
        fn[i] = math.exp( -((x[i]-pos)/(0.6006*wid))**(2*round(n)) )

    return fn

#%%%%%%%%%%

def FouFilter(y,samplingtime,centerfrequency,frequencywidth,shape,mode):
    """Fourier filter function for time-series signal vector y. 

    Keyword arguments:
    'samplingtime' is the total duration of sampled signal in sec, millisec, or microsec;
    'centerfrequency' and 'frequencywidth' are the center frequency and width
    of the filter in Hz, KHz, or MHz, respectively; 'Shape' determines the
    sharpness of the cut-off. If shape = 1, the filter is Gaussian; as
    shape increases the filter shape becomes more and more rectangular.
    Set mode = 0 for band-pass filter, mode = 1 for band-reject (notch) filter.
    FouFilter returns the filtered signal.
    
    T. C. O'Haver (toh@umd.edu),  version 1.5, May, 2007
    Converted to Python by Paul Reasor, May, 2018
    """
    center=centerfrequency*samplingtime  #  center harmonic (fourier component)
    width=frequencywidth*samplingtime    #  width of filter (in harmonics)

    fy=scipy.fftpack.fft(y)  #  Fourier transform of signal
    lft1=range(1,(len(fy)/2)+1)
    lft2=range((len(fy)/2+1),len(fy)+1)

    # Compute filter shape
    ffilter1=ngaussian(lft1,center+1,width,shape)
    ffilter2=ngaussian(lft2,len(fy)-center+1,width,shape)
    ffilter=np.hstack((ffilter1,ffilter2))

    if (mode==1):
        for i in range(len(ffilter)):
            ffilter[i] = 1-ffilter[i]

    if (len(fy)>len(ffilter)): 
        ffilter=np.concatenate((ffilter,ffilter(1)))

    ffy = [0 for j in range(len(fy))] 
   
    for i in range(len(fy)):
        ffy[i]=fy[i]*ffilter[i]  #  Multiply filter by Fourier transform of signal

    return scipy.fftpack.ifft(ffy).real  # Recover filter signal from Fourier transform


#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# END LOCAL FUNCTIONS
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# MAIN MODULE
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# Read in block of rvp8 hdr data for INE and GPS, and determine correction (eps) to be applied to INE data.
# To execute: python make_ine_corrections.py
#
# INE = INE` + eps (INE = raw, INE` = corrected)
# delta = GPS - INE
# Assume: eps = -(delta)_{low-pass-filtered}
# INE` = INE + (-eps)

# Steps preceding execution of this code:
# 
# In /home/hrd
# 0) At beginning of flight remove any existing ine_corrections.dat file
# 1) ls /home/hrd/YYYYMMDDAI/nrs* > /home/hrd/sweeps_list (10 min prior to current time)
# 2) ./dump_headers /home/hrd/sweeps_list

# Path to the rvp8 hrd data extracted and sorted temporally by 'dump_headers_pc.f'
#
path_airs_gps = "/home/sysop/"
file_airs_gps = "rvp8_hdr_dump.dat"

# ray_time gps_lat gps_lon airs_lat airs_lon gps_U gps_V gps_W airs_U airs_V airs_W pitch roll drift gps_age airs_age
#
f = open(path_airs_gps+file_airs_gps,'r')
ntime = file_len(path_airs_gps+file_airs_gps)-1 # Total number of rays (minus 1 for text header)

# If rvp8_hrd_dump file is empty, just exit
#
if (ntime == 0):
    exit()

# Initialize data lists
#
time_airs_gps = [0 for j in range(ntime)]
gps_lat = [0 for j in range(ntime)]
gps_lon = [0 for j in range(ntime)]
airs_lat = [0 for j in range(ntime)]
airs_lon = [0 for j in range(ntime)]
gps_U = [0 for j in range(ntime)]
gps_V = [0 for j in range(ntime)]
airs_U = [0 for j in range(ntime)]
airs_V = [0 for j in range(ntime)]

# Read first-line header
#
line = f.readline() 
line = line.strip()

# Now read the data for each ray
#
lineno=0    
for line in f: 

    line = line.strip()
    columns = line.split()

    time_airs_gps[lineno] = float(columns[0])

    if (columns[1] == "-999.0000" or columns[3] == "-999.0000" or \
        columns[5] == "-999.0000" or columns[8] == "-999.0000"): # Check for flagged values
        gps_lat[lineno] = np.nan
        gps_lon[lineno] = np.nan
        airs_lat[lineno] = np.nan
        airs_lon[lineno] = np.nan
        gps_U[lineno] = np.nan
        gps_V[lineno] = np.nan
        airs_U[lineno] = np.nan
        airs_V[lineno] = np.nan
    else:
        gps_lat[lineno] = float(columns[1])
        gps_lon[lineno] = float(columns[2])
        airs_lat[lineno] = float(columns[3])
        airs_lon[lineno] = float(columns[4])
        gps_U[lineno] = float(columns[5])
        gps_V[lineno] = float(columns[6])
        airs_U[lineno] = float(columns[8])
        airs_V[lineno] = float(columns[9]) 

    lineno+=1

f.close()


#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# Compute delta for each position and wind variable
#
lon_diff = map(operator.sub,gps_lon,airs_lon)
lat_diff = map(operator.sub,gps_lat,airs_lat)
U_diff = map(operator.sub,gps_U,airs_U)
V_diff = map(operator.sub,gps_V,airs_V)


#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# Account for occasional repeated ray times (interpolation below will fail otherwise) 
# For the sake of processing, create lists containing unique ray times.
#
time_elapse_airs_gps = [0 for j in range(ntime)]

for i in range(ntime):
    time_elapse_airs_gps[i] = time_airs_gps[i] - time_airs_gps[0]

max_time = max(time_elapse_airs_gps)

time_elapse_airs_gps_unique = [0 for j in range(ntime)]
lat_diff_unique = [0 for j in range(ntime)]
lon_diff_unique = [0 for j in range(ntime)]
U_diff_unique = [0 for j in range(ntime)]
V_diff_unique = [0 for j in range(ntime)]

count_unique = 0

time_elapse_airs_gps_unique[0] = time_elapse_airs_gps[0]
lat_diff_unique[0] = lat_diff[0]
lon_diff_unique[0] = lon_diff[0]
U_diff_unique[0] = U_diff[0]
V_diff_unique[0] = V_diff[0]

for count in range(1,ntime):

    if (time_elapse_airs_gps[count] != time_elapse_airs_gps[count-1]):

        count_unique+=1
        time_elapse_airs_gps_unique[count_unique] = time_elapse_airs_gps[count]
        lat_diff_unique[count_unique] = lat_diff[count]
        lon_diff_unique[count_unique] = lon_diff[count]
        U_diff_unique[count_unique] = U_diff[count]
        V_diff_unique[count_unique] = V_diff[count]

ntime_unique = count_unique+1  # total number of unique time stamps in file


#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# Linearly interpolate delta to regular 1-sec interval grid
#
dt = 1 # sec

time_1sec = range(0,int(math.floor(max_time))+1,dt)  # Interpolation falls completely within time range
ntime_1sec = len(time_1sec)

ff = interp1d(time_elapse_airs_gps_unique[0:ntime_unique],lat_diff_unique[0:ntime_unique],kind='linear')
lat_diff_1sec = ff(time_1sec)

ff = interp1d(time_elapse_airs_gps_unique[0:ntime_unique],lon_diff_unique[0:ntime_unique],kind='linear')
lon_diff_1sec = ff(time_1sec)

ff = interp1d(time_elapse_airs_gps_unique[0:ntime_unique],U_diff_unique[0:ntime_unique],kind='linear')
U_diff_1sec = ff(time_1sec)

ff = interp1d(time_elapse_airs_gps_unique[0:ntime_unique],V_diff_unique[0:ntime_unique],kind='linear')
V_diff_1sec = ff(time_1sec)


#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# Fill NaNs in delta with average of surrounding data (won't work for NaNs adjacent to one another)
#
for count in range(1,ntime_1sec-1):

    if (np.isnan(lat_diff_1sec[count])):
        lat_diff_1sec[count] = (lat_diff_1sec[count+1]+lat_diff_1sec[count-1])/2.
        lon_diff_1sec[count] = (lon_diff_1sec[count+1]+lon_diff_1sec[count-1])/2.
        U_diff_1sec[count] = (U_diff_1sec[count+1]+U_diff_1sec[count-1])/2.
        V_diff_1sec[count] = (V_diff_1sec[count+1]+V_diff_1sec[count-1])/2.


#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# Fill in remaining NaNs in delta with nearest value to the left (will remove all remaining NaNs)
#
for count in range(1,ntime_1sec):

    if (np.isnan(lat_diff_1sec[count])):
        lat_diff_1sec[count] = lat_diff_1sec[count-1]
        lon_diff_1sec[count] = lon_diff_1sec[count-1]
        U_diff_1sec[count] = U_diff_1sec[count-1]
        V_diff_1sec[count] = V_diff_1sec[count-1]


#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# Remove linear trend in delta (ie, the drift)
#
lat_diff_detrended = detrend(lat_diff_1sec,type='linear')
lon_diff_detrended = detrend(lon_diff_1sec,type='linear')
U_diff_detrended = detrend(U_diff_1sec,type='linear')
V_diff_detrended = detrend(V_diff_1sec,type='linear')


#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# Low-pass filter detrended delta (ie, Schuler oscillation)
# Use central freq of 0, cutoff width of 0.0025 Hz (~6.7 min), Gaussian shape (=2)
#
lat_diff_filter = FouFilter(lat_diff_detrended,time_1sec[ntime_1sec-1],0.0000,.0025,2,0)
lon_diff_filter = FouFilter(lon_diff_detrended,time_1sec[ntime_1sec-1],0.0000,.0025,2,0)
U_diff_filter = FouFilter(U_diff_detrended,time_1sec[ntime_1sec-1],0.0000,.0025,2,0)
V_diff_filter = FouFilter(V_diff_detrended,time_1sec[ntime_1sec-1],0.0000,.0025,2,0)


#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# Store -eps (meps) at 1-sec interval: oscillation + linear trend.
# Interpolate -eps to original time points corresponding to sorted rays.
# Values beyond 1-sec grid just use last-point data
#
lat_meps_1sec = [0 for j in range(ntime_1sec)]
lon_meps_1sec = [0 for j in range(ntime_1sec)]
U_meps_1sec = [0 for j in range(ntime_1sec)]
V_meps_1sec = [0 for j in range(ntime_1sec)]

for i in range(ntime_1sec):
    lat_meps_1sec[i] = lat_diff_filter[i] + (lat_diff_1sec[i]-lat_diff_detrended[i])
    lon_meps_1sec[i] = lon_diff_filter[i] + (lon_diff_1sec[i]-lon_diff_detrended[i])
    U_meps_1sec[i] = U_diff_filter[i] + (U_diff_1sec[i]-U_diff_detrended[i])
    V_meps_1sec[i] = V_diff_filter[i] + (V_diff_1sec[i]-V_diff_detrended[i])

ff = interp1d(time_1sec,lat_meps_1sec,kind='linear',bounds_error=False,fill_value=lat_meps_1sec[ntime_1sec-1])
lat_meps = ff(time_elapse_airs_gps)

ff = interp1d(time_1sec,lon_meps_1sec,kind='linear',bounds_error=False,fill_value=lon_meps_1sec[ntime_1sec-1])
lon_meps = ff(time_elapse_airs_gps)

ff = interp1d(time_1sec,U_meps_1sec,kind='linear',bounds_error=False,fill_value=U_meps_1sec[ntime_1sec-1])
U_meps = ff(time_elapse_airs_gps)

ff = interp1d(time_1sec,V_meps_1sec,kind='linear',bounds_error=False,fill_value=V_meps_1sec[ntime_1sec-1])
V_meps = ff(time_elapse_airs_gps)


#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# Write (append) ray_time airs_lat_corr airs_lon_corr airs_U_corr airs_V_corr
#
out_file = open("/home/sysop/ine_corrections.dat", "a")  # Corrections file for a given flight

#out_file.write("%s\n" % "ray_time, airs_lat_corr, airs_lon_corr, airs_vel_e_corr, airs_vel_n_corr")
for count in range(ntime):
    out_file.write("%9.3f %9.5f %9.5f %9.5f %9.5f\n" % (time_airs_gps[count], lat_meps[count], lon_meps[count], \
          U_meps[count], V_meps[count]))

out_file.close()


#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# Plot raw difference (black) and correction to be applied (red) for U and V
#
time_1sec_hrs = [0 for j in range(ntime_1sec)]

for i in range(ntime_1sec):
    time_1sec_hrs[i] = time_1sec[i]/3600.

figa = plt.figure(figsize=(19.5,12))

fig  = figa.add_subplot(211)
fig2 = plt.plot(time_1sec_hrs, U_diff_1sec, 'k-', time_1sec_hrs, U_meps_1sec, 'r-')

plottitle = "GPS-INE"
fig.set_title(plottitle,fontsize=18)
fig.set_ylabel("U (m/s)",fontsize=18)
fig.set_xlim([min(time_1sec_hrs),max(time_1sec_hrs)])
fig.set_ylim([-10,10])

fig = figa.add_subplot(212)
fig2 = plt.plot(time_1sec_hrs, V_diff_1sec, 'k-', time_1sec_hrs, V_meps_1sec, 'r-')

fig.set_xlabel("Elapsed Time (h)",fontsize=18)
fig.set_ylabel("V (m/s)",fontsize=18)
fig.set_xlim([min(time_1sec_hrs),max(time_1sec_hrs)])
fig.set_ylim([-10,10])

figa.subplots_adjust(wspace=.25)

#plt.show()

####### Save Figure to File

figfile = "/home/sysop/ine_corrections.png"
plt.savefig(figfile, bbox_inches='tight')

plt.clf()

