/* These functions use Xlib functions to plot arrows pointing with wind
 *  or meteorological wind barbs. Translated from Fortran Plot10 subs.
 * To preserve the right-handed math i use JFLIP to reverse direction for
 * dy, because X assumes 0,0 is at upper left, while "normal"
 * algebra and trig assume 0,0 at lower left. NOTE that these routines are
 * passed coordinates in the X system, so they need not be adjusted.
 */
# include <X11/Intrinsic.h>
# include <X11/Xlib.h>
# include <X11/Xutil.h> 
# include <math.h>

# define DEG2RAD 0.0174533
# define JFLIP -1

int xplot_arrow(display, drawable, gc,
		x, y, shaft, u, v)
	Display *display;
	Drawable drawable;	
	GC gc;
	Position x, y;
	double shaft, u, v;
	
/* x,y is end and shaft is length, in pixels, of arrow shaft.
 * u, v in m/s */

{
	Position xp, yp, xp2, yp2;
	double spd, xe, ye, dx, dy, th, theta, edge;
	
	spd = sqrt(u*u + v*v);
	if( spd > 0.0 ) {
		dx = shaft * u / spd;
		dy = (shaft * v / spd) * JFLIP;
		xe = x + dx;
		ye = y + dy;
		xp = xe + 0.5;
		yp = ye + 0.5;
		XDrawLine(display, drawable, gc, x, y, xp, yp);
/* the arrow head,  arrow edge scaled by length of shaft*/
		edge = sqrt (dx*dx + dy*dy) * 0.2;
		theta = atan2(-dy,-dx);
/* draw edges 30 degrees from point */
		th = theta + .5235;	 /* +30 deg */
		xp2 = xe + edge*cos(th) + 0.5;
		yp2 = ye + edge*sin(th) + 0.5;
		XDrawLine(display, drawable, gc, xp, yp, xp2, yp2);
		th = theta - .5235; 	/* -30 deg */
		xp2 = xe + edge*cos(th) + 0.5;
		yp2 = ye + edge*sin(th) + 0.5;
		XDrawLine(display, drawable, gc, xp, yp, xp2, yp2);
	}
	return 0;
}
/* -------------------------------------- */
int xplot_barb_wd(display, drawable, gc,xc,yc,dir,spd,shaft)

	Display *display;
	Drawable drawable;	
	GC gc;
	Position xc, yc;
	double dir, spd, shaft;
/* Draw a wind barb from xc, yc that is shaft pixels long
 * given speeds in m/s, and dir is met angle in degrees. */
{
	Position xp, yp, xp2, yp2;
	double xe, ye, ang, slant, vel,  qos, syn, qosp, synp;
	double c_slant, s_slant, dr, dx, dy, barb_len;
	int iflag;
 
	slant = 1.22713;	/* 70 deg in radians, slant for barbs */
	ang = (450.0-dir)*DEG2RAD; /* met to math angle */
	vel = spd * 2.0; /* m/s to knots */
	if(vel <= 0.0)	/* then bad value so do not plot... */
              return -1;

	qos = cos(ang);
	syn = sin(ang);
	c_slant = cos(slant);
	s_slant = sin(slant);
	dr = shaft * 0.1;		/* spacing along shaft */
	dx = dr * qos;
	dy = dr * syn * JFLIP;
	barb_len = shaft * 0.25;
	qosp = barb_len * (qos * c_slant + syn * s_slant); /* cos (ang-slant) */
	synp = barb_len * (syn * c_slant - qos * s_slant); /* sin (ang-slant) */
	xe = xc + shaft * qos;
	ye = yc + shaft * syn * JFLIP;
	xp = xe + 0.5;
	yp = ye + 0.5;
	
	XDrawLine(display, drawable, gc, xc, yc, xp, yp); 

	iflag = 0;
	while (vel >= 48.0){  /* draw flag  */
		iflag = 1;
              	xp2 = xe+qosp + 0.5;
              	yp2 = ye+synp*JFLIP + 0.5;
		XDrawLine(display, drawable, gc, xp, yp, xp2, yp2); 
              	xe = xe-dx;
              	ye = ye-dy;
              	xp = xe + 0.5;
              	yp = ye + 0.5;
		XDrawLine(display, drawable, gc, xp, yp, xp2, yp2); 
              	vel = vel-50.0;
	}
	
	while (vel >= 8.0) {	/* draw barb */
		iflag = 1;
		xp2 = xe+qosp + 0.5;
              	yp2 = ye+synp*JFLIP + 0.5;
		XDrawLine(display, drawable, gc, xp, yp, xp2, yp2); 
              	xe = xe-dx;
              	ye = ye-dy;
              	xp = xe + 0.5;
              	yp = ye + 0.5;
		vel = vel - 10.0;
	}
	
	if(vel >= 3.0){ 		/* draw half barb */
		if(iflag == 0){ 	/* move down from end of shaft */
              		xe = xe - dx;
              		ye = ye - dy;
			xp = xe + 0.5;
              		yp = ye + 0.5;
		}
        	xp2 = xe+0.5*qosp + 0.5;
              	yp2 = ye+0.5*synp *JFLIP + 0.5;
		XDrawLine(display, drawable, gc, xp, yp, xp2, yp2); 
	}
	return 0;
}

/*		 -------------------		*/

int xplot_barb_uv(display, drawable, gc,xc,yc,u,v,shaft)
	Display *display;
	Drawable drawable;	
	GC gc;
	Position xc, yc;
	float shaft, u, v;
/* Draw a wind barb from xc, yc that is shaft pixels long
 *  given u, v in m/s,
 */
{
	Position xp, yp, xp2, yp2;
	float ang, slant, vel, xe, ye, qos, syn, qosp, synp;
	double c_slant, s_slant, dr, dx, dy, barb_len;
	int iflag;

	vel = sqrt(u*u + v*v) * 2.0; /* m/s to knots */
	if(vel <= 0.0)
              return -1;
	ang = atan2(-v, -u);  /* angle of flag blowing with the wind*/
	slant = 1.22713;	/* 70 deg in radians, slant for barbs */

	qos = cos(ang);
	syn = sin(ang);
	c_slant = cos(slant);
	s_slant = sin(slant);
	dr = shaft * 0.1;		/* spacing along shaft */
	dx = dr * qos;
	dy = dr * syn * JFLIP;
	barb_len = shaft * 0.25;
	qosp = barb_len * (qos * c_slant + syn * s_slant); /* cos (ang-slant) */
	synp = barb_len * (syn * c_slant - qos * s_slant); /* sin (ang-slant) */
	xe = xc + shaft * qos;
	ye = yc + shaft * syn * JFLIP;
	xp = xe + 0.5;
	yp = ye + 0.5;
	
	XDrawLine(display, drawable, gc, xc, yc, xp, yp); 

	iflag = 0;
	while (vel >= 48.0){  /* draw flag  */
		iflag = 1;
              	xp2 = xe+qosp + 0.5;
              	yp2 = ye+synp*JFLIP + 0.5;
		XDrawLine(display, drawable, gc, xp, yp, xp2, yp2); 
              	xe = xe-dx;
              	ye = ye-dy;
              	xp = xe + 0.5;
              	yp = ye + 0.5;
		XDrawLine(display, drawable, gc, xp, yp, xp2, yp2); 
              	vel = vel-50.0;
	}
	
	while (vel >= 8.0) {	/* draw barb */
		iflag = 1;
		xp2 = xe+qosp + 0.5;
              	yp2 = ye+synp*JFLIP + 0.5;
		XDrawLine(display, drawable, gc, xp, yp, xp2, yp2); 
              	xe = xe-dx;
              	ye = ye-dy;
              	xp = xe + 0.5;
              	yp = ye + 0.5;
		vel = vel - 10.0;
	}
	
	if(vel >= 3.0){ 		/* draw half barb */
		if(iflag == 0){ 	/* move down from end of shaft */
              		xe = xe - dx;
              		ye = ye - dy;
			xp = xe + 0.5;
              		yp = ye + 0.5;
		}
        	xp2 = xe+0.5*qosp + 0.5;
              	yp2 = ye+0.5*synp *JFLIP + 0.5;
		XDrawLine(display, drawable, gc, xp, yp, xp2, yp2); 
	}
return 0;
}
